/*
 *  Copyright (C) 2004 Christophe Fergeau
 *
 * 
 *  The code contained in this file is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either version
 *  2.1 of the License, or (at your option) any later version.
 *  
 *  This file is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this code; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *  iTunes and iPod are trademarks of Apple
 * 
 *  This product is not supported/written/published by Apple!
 *
 */

#define FINGERPRINT_FILE "fingerprints.xml"

#include <libxml/xmlreader.h>
#include <glib.h>
#include <string.h>
#include <errno.h>

#include "rb-fingerprints.h"

static int
read_next (xmlTextReader *reader)
{
	int ret;
	int depth;

	depth = xmlTextReaderDepth (reader);
	
	ret = xmlTextReaderRead (reader);
	while (ret == 1) {
		if (xmlTextReaderDepth (reader) == depth) {
			return 1;
		} else if (xmlTextReaderDepth (reader) < depth) {
			return 0;
		}
		ret = xmlTextReaderRead (reader);
	}

	return ret;
}

static char *
read_string_value (xmlTextReader *reader)
{
	int ret;
	char *text;

	text = NULL;	
	ret = xmlTextReaderRead (reader);
	while (ret == 1) {
		xmlReaderTypes type;
		
		type = xmlTextReaderNodeType (reader);
		if (type == XML_READER_TYPE_TEXT) {
			g_free (text);
			text = g_strdup (xmlTextReaderConstValue (reader));
		}

		ret = read_next (reader);
	}

	return text;
}

static void 
process_audio_fingerprint (xmlTextReader *reader, GHashTable *fingerprints)
{
	int ret;
	char *id;
	char *fingerprint;

	if (strcmp (xmlTextReaderConstName (reader), "audio-fingerprint")) {
		g_print ("Found %s instead of audio-fingerprint\n", 
			 xmlTextReaderConstName (reader));
		return;
	}

	id = NULL;
	fingerprint = NULL;
	ret = xmlTextReaderRead (reader);
	while (ret == 1) {
		xmlReaderTypes type;
		
		type = xmlTextReaderNodeType (reader);

		if (type == XML_READER_TYPE_ELEMENT) {
			const char *name;
			
			name = xmlTextReaderConstName (reader);
			if (!strcmp (name, "fingerprint")) {
				g_free (fingerprint);
				fingerprint = read_string_value (reader);
			} else if (!strcmp (name, "file-id")) {
				g_free (id);
				id = read_string_value (reader);
			}
		}

		ret = read_next (reader);
	}

	if ((id != NULL) && (fingerprint != NULL)) {
		g_hash_table_insert (fingerprints, id, fingerprint);
	} else {
		g_free (id);
		g_free (fingerprint);
	}
}

static GHashTable *
process_audio_fingerprints (xmlTextReader *reader)
{
	int ret;
	GHashTable *fingerprints;

	if (strcmp (xmlTextReaderConstName (reader), "audio-fingerprints")) {
		g_print ("Found %s instead of audio-fingerprints\n", 
			 xmlTextReaderConstName (reader));
		return NULL;
	}

	fingerprints = g_hash_table_new_full (g_str_hash, g_str_equal, 
					      g_free, g_free);

	ret = xmlTextReaderRead (reader);
	while (ret == 1) {
		xmlReaderTypes type;

		type = xmlTextReaderNodeType (reader);
		if (type == XML_READER_TYPE_ELEMENT) {
			process_audio_fingerprint (reader, fingerprints);
		}
		ret = read_next (reader);
	}

	return fingerprints;
}


static GHashTable *
generic_fingerprints_load (const char *filename)
{      
	xmlTextReader *reader;
	int ret;
	GHashTable *fingerprints;

	LIBXML_TEST_VERSION;

	reader = xmlReaderForFile (filename, NULL, 0);
	if (reader == NULL) {
		return NULL;
	}

	fingerprints = NULL;
	ret = xmlTextReaderRead (reader);
	while ((ret == 1) && (fingerprints == NULL)) {
		xmlReaderTypes type;

		type = xmlTextReaderNodeType (reader);

		if (type == XML_READER_TYPE_ELEMENT) {
			fingerprints = process_audio_fingerprints (reader);
		}
		ret = read_next (reader);
	}

	xmlFreeTextReader (reader);

	return fingerprints;
}


GHashTable *
rb_fingerprints_load (const char *rb_dot_dir)
{
	GHashTable *fingerprints;
	char *filename;

	filename = g_build_filename (rb_dot_dir, FINGERPRINT_FILE, NULL);
	fingerprints = generic_fingerprints_load (filename);
	g_free (filename);
	
	return fingerprints;
}


static void
write_fingerprint (FILE *f, const char *fingerprint, const char *file_id)
{
	char *encoded;

	fprintf (f, "<audio-fingerprint>");
	encoded = xmlEncodeEntitiesReentrant (NULL, file_id);
	fprintf (f, "<file-id>%s</file-id>", encoded);
	g_free (encoded);

	encoded = xmlEncodeEntitiesReentrant (NULL, fingerprint);
	fprintf (f, "<fingerprint>%s</fingerprint>", encoded);
	g_free (encoded);
	fprintf (f, "</audio-fingerprint>");
}

typedef char *(*GetIdFunc)(iPodSong *song, gpointer data);

static char *
get_rb_id (iPodSong *song, gpointer data)
{
	return g_strdup (song->pc_path_locale);
}


static int
generic_fingerprints_save (const char *filename, GList *songs, 
			   GetIdFunc get_file_id, gpointer data)
{
	FILE *f;
	GList *s;
	f = fopen (filename, "w");
	if (f == NULL) {
		return errno;
	}
	
	fprintf (f, "%s\n%s\n", "<?xml version=\"1.0\" standalone=\"yes\"?>",
		 "<audio-fingerprints version=\"1.0\">");

	for (s = songs; s != NULL; s = s->next) {
		iPodSong *song;

		song = s->data;
		if (song->fingerprint != NULL) {
			char *id;

			id = get_file_id (song, data);
			write_fingerprint (f, song->fingerprint, id);
			g_free (id);
		}
	}

	fprintf (f, "</audio-fingerprints>");
	fclose (f);

	return 0;
}


int
rb_fingerprints_save (const char *rb_dot_dir, iPodDB *db)
{
	gchar *filename;
	GList *songs;
	
	filename = g_build_filename (rb_dot_dir, FINGERPRINT_FILE, NULL);
	songs = ipod_db_get_songs (db);
	generic_fingerprints_save (filename, songs, get_rb_id, NULL);
	g_list_foreach (songs, (GFunc)ipod_song_unref, NULL);
	g_list_free (songs);
	g_free (filename);
}

