/*
 *  Copyright (C) 2004 Christophe Fergeau
 *
 * 
 *  The code contained in this file is free software; you can redistribute
 *  it and/or modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either version
 *  2.1 of the License, or (at your option) any later version.
 *  
 *  This file is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this code; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <config.h>
#include <glib.h>
#include <glib/gi18n.h>
#include <libitunesdb/ipoddb.h>
#include <libgnomevfs/gnome-vfs-xfer.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <locale.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <libitunesdb/ipod-detect.h>

#include "rhythmdb-parse.h"
#include "rb-fingerprints.h"
#include "gst-md5.h"


static void
song_set_fingerprint (iPodDB *db, iPodSong *song)
{
	gchar *uri;
	gchar *md5;

	if (song->ipod_path != NULL) {
		gchar *mount_point;
		gchar *unix_path;

		mount_point = ipod_db_get_mount_point (db);
		unix_path = ipod_path_to_unix_path (mount_point, 
						    song->ipod_path);
		uri = gnome_vfs_make_uri_canonical (unix_path);
		g_free (mount_point);
		g_free (unix_path);
	} else {
		g_assert (song->pc_path_locale != NULL);       
		uri = gnome_vfs_make_uri_canonical (song->pc_path_locale);
	}

	md5 = md5_generator_get_md5 (uri);
	g_free (uri);
	ipod_db_set_song_fingerprint (db, song, md5);
	g_free (md5);
}


static int
copy_to_ipod (iPodDB *db, iPodSong *song)
{
	int res;
	char *mount_point;
	char *ipod_path;

	mount_point = ipod_db_get_mount_point (db);
	res = ipod_transfer_uri (mount_point, song->pc_path_locale, 
				 &ipod_path);

	g_free (song->ipod_path);
	if (ipod_path != NULL) {
		song->ipod_path = ipod_path_from_unix_path (mount_point, 
							    ipod_path);	
	} else {
		song->ipod_path = NULL;
	}
	g_free (mount_point);

	if (res == 0) {
		gchar *display_name;
		gchar *src_path;
		gchar *dest_path;
		gchar *filename;

		song->transferred = TRUE;

		/* Set song fingerprint to be able to handle to detect
		 * that the song has already been transferred in subsequent
		 * run of this program.
		 */
		if (song->fingerprint == NULL) {
			song_set_fingerprint (db, song);
		}
		display_name = gnome_vfs_format_uri_for_display (song->pc_path_locale);
		filename = g_path_get_basename (display_name);
		src_path = g_path_get_dirname (display_name);
		g_free (display_name);
		dest_path = g_path_get_dirname (ipod_path);
		printf ("Copied %s\n\tfrom %s\n\tto %s\n\n",
			filename, src_path, dest_path);
		g_free (filename);
		g_free (src_path);
		g_free (dest_path);

		res = ipod_db_add_song (db, song);
		if (res != 0) {
			/* We failed to add the file to the database, don't
			 * leave "orphan" files lying around in the directory
			 * structure
			 */
			char *mount_point;
			char *unix_path;

			mount_point = ipod_db_get_mount_point (db);
			unix_path = ipod_path_to_unix_path (mount_point, 
							    song->ipod_path);
			gnome_vfs_unlink (unix_path);
			g_free (mount_point);
			g_free (unix_path);
		}
	} else {
		g_warning ("Error transferring %s\n", song->pc_path_locale);
	}

	g_free (ipod_path);

	return res;
}


static int
transfer_songs_to_ipod (iPodDB *db, GList *songs)
{
	GList *s;
	int result;
	GList *invalid_songs;

	invalid_songs = NULL;

	for (s = songs ; s != NULL; s = s->next) {
		iPodSong *song;	
		
		song = (iPodSong *)s->data;
		if (song->fingerprint == NULL) {
			/* Don't transfer songs with NULL fingerprints, 
			 * if we have such songs this means there is
			 * another song in the database with the same
			 * fingerprint.
			 */
			invalid_songs = g_list_append (invalid_songs, song);
		} else if (song->transferred == FALSE) {
			result = copy_to_ipod (db, song);
			if (result != 0) {
				invalid_songs = g_list_append (invalid_songs, 
							       song);
			}
		}
	}

	/* FIXME: do something with the invalid songs ;) */
	if (invalid_songs != NULL) {
		g_list_free (invalid_songs);
		return -1;
	}

	return 0;
}

static void
db_set_fingerprints (iPodDB *db)
{
	GList *songs;
	GList *s;

	songs = ipod_db_get_songs (db);
	for (s = songs; s != NULL; s = s->next) {
		iPodSong *song;

		song = (iPodSong*)s->data;
		if (song->fingerprint == NULL) {
			song_set_fingerprint (db, song);
/*			g_print ("%s fingerprint is %s **ADDED**\n", 
			song->pc_path_locale, song->fingerprint);*/
		} else {
/*			g_print ("%s fingerprint is %s\n", 
			song->pc_path_locale, song->fingerprint);*/
		}
	}
	g_list_foreach (songs, (GFunc)ipod_song_unref, NULL);
	g_list_free (songs);
}

static void
rhythmdb_set_fingerprints (const char *rb_dir, iPodDB *db)
{
	GHashTable *fingerprints;
	GList *songs;
	GList *s;

	fingerprints = rb_fingerprints_load (rb_dir);
	if (fingerprints != NULL) {
		songs = ipod_db_get_songs (db);
		for (s = songs; s != NULL; s = s->next) {
			iPodSong *song;
			gchar *fingerprint;

			song = (iPodSong*)s->data;
			fingerprint = g_hash_table_lookup (fingerprints, 
							   song->pc_path_locale);
			if (fingerprint != NULL) {
				ipod_db_set_song_fingerprint (db, song, 
							      fingerprint);
			}
		}
		g_list_foreach (songs, (GFunc)ipod_song_unref, NULL);
		g_list_free (songs);
		g_hash_table_destroy (fingerprints);
	}
	db_set_fingerprints (db);
}


static iPodDB *
load_ipod_db (const char *mount_point)
{
	iPodDB *db;

	db = ipod_db_load (mount_point);
	if (db == NULL) {
		return NULL;
	}

	db_set_fingerprints (db);

	return db;
}

static gboolean
has_song_with_fingerprint (iPodDB *db, const char *fingerprint)
{
	iPodSong *song;

	song = ipod_db_get_song_by_fingerprint (db, fingerprint);
	if (song != NULL) {
		ipod_song_unref (song);
	}

	return (song != NULL);
}

static GList *
db_diff (iPodDB *db1, iPodDB *db2)
{
	GList *new_songs;
	GList *songs;
	GList *s;

	songs = ipod_db_get_songs (db2);
	
	new_songs = NULL;
	for (s = songs; s != NULL; s = s->next) {
		iPodSong *song;
		

		song = (iPodSong *)s->data;
		/* songs can have a NULL fingerprint if we have several songs 
		 * with the same hash, the 1st one will have a non-NULL 
		 * fingerprint and all the others will have a NULL fingerprint.
		 */
		if (song->fingerprint == NULL) {
/*			g_print ("No fingerprint, adding to new_songs list\n");*/
		}
		if (!has_song_with_fingerprint (db1, song->fingerprint)) {
/*			g_print ("No song with %s fingerprint in %x, adding to new_songs list\n", song->fingerprint, (int)db1);*/
		}
		if ((song->fingerprint == NULL) || (!has_song_with_fingerprint (db1, song->fingerprint))) {
			ipod_song_ref (song);
			new_songs = g_list_prepend (new_songs, song);
		}
	}

	g_list_foreach (songs, (GFunc)ipod_song_unref, NULL);
	g_list_free (songs);

	return new_songs;
}

static GList *
get_songs_to_add (iPodDB *db, iPodDB *to_merge)
{
	GList *songs;
	GList *s;
	
	songs = db_diff (db, to_merge);

	for (s = songs; s != NULL; s = s->next) {
		iPodSong *song;

		song = (iPodSong *)s->data;
		g_print ("\t%s - %s - %s\n", song->artist, song->album,
			 song->title);
	}

	return songs;
}

static GList *
get_songs_to_remove (iPodDB *db, iPodDB *to_merge)
{
	GList *songs;
	GList *s;

	songs = db_diff (to_merge, db);

	for (s = songs; s != NULL; s = s->next) {
		iPodSong *song;
				
		song = (iPodSong *)s->data;		
		g_print ("\t%s - %s - %s\n", song->artist, song->album,
			 song->title);
	}

	return songs;
}

char *
rb_dot_dir (void)
{
	return g_build_filename (g_get_home_dir (), ".gnome2",
				 "rhythmbox", NULL);
}

static gchar *
get_ipod_path (void)
{
	gchar *res;
	GList *ipods;

	ipods = ipod_enumerate ();
	if (ipods == NULL) {
		return NULL;
	}

	res = g_strdup (ipods->data);
	g_list_foreach (ipods, (GFunc)g_free, NULL);
	g_list_free (ipods);

	return res;
}

static int
do_add_new_songs (iPodDB *ipod_db, iPodDB *rb_db)
{
	GList *to_copy;
	int result;

	g_print ("**** Adding new songs to the iPod ****\n");
	to_copy = get_songs_to_add (ipod_db, rb_db);
	g_print ("\n");

	result = 0;
	if (to_copy != NULL) {
		g_print ("**** Transferring new songs to the iPod ****\n");
		result = transfer_songs_to_ipod (ipod_db, to_copy);
		g_print ("\n");

		if (result == 0) {
			printf ("%d new files successfully transferred to iPod\n", g_list_length (to_copy));
		} else {
			printf ("Error while transferring files to iPod\n");
		}

		g_list_foreach (to_copy, (GFunc)ipod_song_unref, NULL);
		g_list_free (to_copy);		
	} else {
		g_print ("No new songs to transfer to the iPod\n");
	}
	g_print ("\n");

	return result;
}

static int
do_remove_deleted_songs (iPodDB *ipod_db, iPodDB *rb_db)
{
	GList *to_remove;

	g_print ("**** Removing old songs from the iPod ****\n");
	to_remove = get_songs_to_remove (ipod_db, rb_db);
	g_print ("\n");
	
	if (to_remove != NULL) {
		GList *s;
		
		for (s = to_remove; s != NULL; s = s->next) {
			iPodSong *song;
			gchar *ipod_path;
			gchar *mount_point;

			song = (iPodSong *)s->data;
			ipod_db_remove_song (ipod_db, song);

			mount_point = ipod_db_get_mount_point (ipod_db);
			ipod_path = ipod_path_to_unix_path (mount_point, 
							    song->ipod_path);
			g_assert (ipod_path != NULL);
			if (song->pc_path_locale != NULL) {
				gnome_vfs_unlink (ipod_path);
			}
			g_free (mount_point);
			g_free (ipod_path);
		}
		g_list_foreach (to_remove, (GFunc)ipod_song_unref, NULL);
		g_list_free (to_remove);
	} else {
		g_print ("No songs to remove from the iPod\n");
	}
	
	return 0;
}

static gint 
match_pl_name (gconstpointer a, gconstpointer b)
{
	iPodPlaylist *pl1;
	iPodPlaylist *pl2;
	
	pl1 = (iPodPlaylist *)a;
	pl2 = (iPodPlaylist *)b;

	return (strcmp (pl1->name, pl2->name));
}

static void
do_merge_playlists (iPodDB *ipod_db, iPodDB *rb_db)
{
	GList *ipod_pl;
	GList *rb_pl;
	GList *pl;

	ipod_pl = ipod_db_get_playlists (ipod_db);
	rb_pl = ipod_db_get_playlists (rb_db);

	/* Remove playlists that are on the iPod but not in rb */
	for (pl = ipod_pl; pl != NULL; pl = pl->next) {
		GList *playlist;
		playlist = g_list_find_custom (rb_pl, NULL, match_pl_name);
		if (playlist == NULL) {
			ipod_db_remove_playlist (ipod_db, pl->data);
		}
	}

	/* Add playlists that are in rb but not on the iPod */
	for (pl = rb_pl; pl != NULL; pl = pl->next) {
		GList *playlist;
		playlist = g_list_find_custom (ipod_pl, NULL, match_pl_name);
		if (playlist == NULL) {
			ipod_db_add_playlist (ipod_db, pl->data);
		}
	}

	g_list_foreach (ipod_pl, (GFunc)ipod_playlist_unref, NULL);
	g_list_foreach (rb_pl, (GFunc)ipod_playlist_unref, NULL);
	g_list_free (ipod_pl);
	g_list_free (rb_pl);
}

static void
do_ipod_rb_sync (iPodDB *ipod_db, iPodDB *rb_db)
{
	do_add_new_songs (ipod_db, rb_db);
	do_remove_deleted_songs (ipod_db, rb_db);

	do_merge_playlists (ipod_db, rb_db);
	ipod_db_save (ipod_db);
}


static iPodDB *
load_rb_db (const char *ipod_path)
{
	iPodDB *rb_db;
	gchar *rb_dir;
	gchar *filename;

	g_print ("**** Loading rhythmbox song library ****\n\n");
	rb_db = ipod_db_new (ipod_path);

	if (rb_db == NULL) {
		return NULL;
	}

	rb_dir = rb_dot_dir ();

	filename = g_build_filename (rb_dir, "rhythmdb.xml", NULL);
	rhythmdb_tree_load (filename, rb_db);
	g_free (filename);

	g_print ("**** Setting fingerprints for rhythmbox songs ****\n\n");
	rhythmdb_set_fingerprints (rb_dir, rb_db);
	rb_fingerprints_save (rb_dir, rb_db);

	g_print ("**** Reading rhythmbox playlists ****\n\n");
	filename = g_build_filename (rb_dir, "playlists.xml", NULL);
	rb_load_playlists (rb_db, filename);
	g_free (filename);
	g_free (rb_dir);

	return rb_db;
}


int
main (int argc, char **argv)
{
	char *dest;
	iPodDB *rb_db;
	iPodDB *ipod_db;
	struct timeval tv;
	gchar *ipod_path;

	bindtextdomain (GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
        bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
        textdomain (GETTEXT_PACKAGE);
	setlocale (LC_ALL, "");

	gettimeofday (&tv, NULL);
	g_random_set_seed (tv.tv_usec);

	gnome_vfs_init ();
	gst_init (&argc, &argv);

	ipod_path = get_ipod_path ();       
	if (ipod_path == NULL) {
		g_print ("Couldn't find a mounted iPod\n");
		return -1;
	}

	g_print ("**** Loading database from the iPod ****\n\n");
	ipod_db = load_ipod_db (ipod_path);
	if (ipod_db == NULL) {
		g_print ("Couldn't load iPod database, exiting.\n");
		exit (-1);
	}

	rb_db = load_rb_db (ipod_path);
	g_free (ipod_path);
	if (rb_db == NULL) {
		g_print ("Couldn't create iPod database, exiting.\n");
		exit (-1);
	}

	do_ipod_rb_sync (ipod_db, rb_db);

	ipod_db_destroy (ipod_db);
	ipod_db_destroy (rb_db);

	gnome_vfs_shutdown ();

	return 0;
}
