/*
 *  Copyright (C) 2003, 2004 Colin Walters <walters@verbum.org>
 *  Copyright (C) 2004 Christophe Fergeau  
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#include <glib.h>
#include <libgnomevfs/gnome-vfs-utils.h>
#include <libitunesdb/ipoddb.h>
#include <libxml/entities.h>
#include <libxml/SAX.h>
#include <libxml/parserInternals.h>
#include <string.h>

#include "rhythmdb-parse.h"

//#define D(x...) g_print (x)
#define D(x...)


typedef enum
{
	RHYTHMDB_PROP_INVALID = -1,
	RHYTHMDB_PROP_TYPE,
	RHYTHMDB_PROP_TITLE,
	RHYTHMDB_PROP_GENRE,
	RHYTHMDB_PROP_ARTIST,
	RHYTHMDB_PROP_ALBUM,
	RHYTHMDB_PROP_TRACK_NUMBER,
	RHYTHMDB_PROP_DISC_NUMBER,
	RHYTHMDB_PROP_DURATION,
	RHYTHMDB_PROP_FILE_SIZE,
	RHYTHMDB_PROP_LOCATION,
	RHYTHMDB_PROP_MTIME,
	RHYTHMDB_PROP_RATING,
	RHYTHMDB_PROP_AUTO_RATE,
	RHYTHMDB_PROP_PLAY_COUNT,
	RHYTHMDB_PROP_LAST_PLAYED,
	RHYTHMDB_PROP_BITRATE,
	RHYTHMDB_PROP_TRACK_GAIN,
	RHYTHMDB_PROP_TRACK_PEAK,
	RHYTHMDB_PROP_ALBUM_GAIN,
	RHYTHMDB_PROP_ALBUM_PEAK,
	RHYTHMDB_PROP_MIMETYPE,

} RhythmDBPropType;

struct _XmlToEntryTypeMap {
	gchar *xml_tag;
	RhythmDBPropType propid;
};

typedef struct _XmlToEntryTypeMap XmlToEntryTypeMap;



static XmlToEntryTypeMap xmlToEntryArray[] = 
{
	{"type",         RHYTHMDB_PROP_TYPE},
	{"title",        RHYTHMDB_PROP_TITLE},
	{"genre",        RHYTHMDB_PROP_GENRE},
	{"artist",       RHYTHMDB_PROP_ARTIST},
	{"album",        RHYTHMDB_PROP_ALBUM},
	{"track-number", RHYTHMDB_PROP_TRACK_NUMBER},
	{"disc-number",  RHYTHMDB_PROP_DISC_NUMBER},
	{"duration",     RHYTHMDB_PROP_DURATION},
	{"file-size",    RHYTHMDB_PROP_FILE_SIZE},
	{"location",     RHYTHMDB_PROP_LOCATION},
	{"mtime",        RHYTHMDB_PROP_MTIME},
	{"rating",       RHYTHMDB_PROP_RATING},
	{"last-played",  RHYTHMDB_PROP_LAST_PLAYED},
	{"bitrate",      RHYTHMDB_PROP_BITRATE},
	{"mimetype",     RHYTHMDB_PROP_MIMETYPE}
};

static RhythmDBPropType
xml_tag_to_entry_type (const gchar *tag)
{
	int i;
	int size = sizeof (xmlToEntryArray)/sizeof (XmlToEntryTypeMap);

	for (i = 0; i < size; i++) {
		if (strcmp (xmlToEntryArray[i].xml_tag, tag) == 0) {
			return xmlToEntryArray[i].propid;
		}
	}
	return RHYTHMDB_PROP_INVALID;
}

static const char *
entry_type_to_xml_tag (RhythmDBPropType type)
{
	int i;
	int size = sizeof (xmlToEntryArray)/sizeof (XmlToEntryTypeMap);

	for (i = 0; i <= size; i++) {
		if (xmlToEntryArray[i].propid == type) {
			return xmlToEntryArray[i].xml_tag;
		}
	}
	return NULL;
}


struct RhythmDBTreeLoadContext
{
	iPodDB *db;
	iPodSong *song;
	xmlParserCtxtPtr xmlctx;
	enum {
		RHYTHMDB_TREE_PARSER_STATE_START,
		RHYTHMDB_TREE_PARSER_STATE_RHYTHMDB,
		RHYTHMDB_TREE_PARSER_STATE_ENTRY,
		RHYTHMDB_TREE_PARSER_STATE_ENTRY_PROPERTY,
		RHYTHMDB_TREE_PARSER_STATE_END,
	} state;
	gboolean in_unknown_elt;
	char *genrename;
	char *albumname;
	char *artistname;
	GString *buf;
	RhythmDBPropType propid;
	gboolean add_song_to_db;
};

static void
rhythmdb_tree_parser_start_element (struct RhythmDBTreeLoadContext *ctx,
				    const char *name, const char **attrs)
{
	if (ctx->in_unknown_elt)
		return;

	switch (ctx->state)
	{
	case RHYTHMDB_TREE_PARSER_STATE_START:
	{
		if (!strcmp (name, "rhythmdb"))
			ctx->state = RHYTHMDB_TREE_PARSER_STATE_RHYTHMDB;
		else
			ctx->in_unknown_elt = TRUE;
		break;
	}
	case RHYTHMDB_TREE_PARSER_STATE_RHYTHMDB:
	{
		if (!strcmp (name, "entry")) {
			gboolean type_set = FALSE;
			for (; *attrs; attrs +=2) {
				if (!strcmp (*attrs, "type")) {
					const char *typename = *(attrs+1);
					if (strcmp (typename, "song"))
						return;
					type_set = TRUE;
					break;
				}
			}
			g_assert (type_set);
			ctx->state = RHYTHMDB_TREE_PARSER_STATE_ENTRY;
			ctx->song = ipod_song_new ();
			ctx->add_song_to_db = FALSE;
		} else
			ctx->in_unknown_elt = TRUE;
		break;
	}
	case RHYTHMDB_TREE_PARSER_STATE_ENTRY:
	{
		int val = xml_tag_to_entry_type (name);
		if (val < 0) {
			ctx->in_unknown_elt = TRUE;
			break;
		}
		
		ctx->state = RHYTHMDB_TREE_PARSER_STATE_ENTRY_PROPERTY;
		ctx->propid = val;
		ctx->buf = g_string_new ("");
		break;
	}
	case RHYTHMDB_TREE_PARSER_STATE_ENTRY_PROPERTY:
	case RHYTHMDB_TREE_PARSER_STATE_END:
	break;
	}
}

static void
rhythmdb_tree_parser_end_element (struct RhythmDBTreeLoadContext *ctx, const char *name)
{
	if (ctx->in_unknown_elt) {
		ctx->in_unknown_elt = FALSE;		
		return;
	}
	
	switch (ctx->state)
	{
	case RHYTHMDB_TREE_PARSER_STATE_RHYTHMDB:
		ctx->state = RHYTHMDB_TREE_PARSER_STATE_END;
		break;
	case RHYTHMDB_TREE_PARSER_STATE_ENTRY:
	{
		/* Entry is complete */
		if (ctx->add_song_to_db) {
			ipod_db_add_song (ctx->db, ctx->song);
		}
		ipod_song_unref (ctx->song);
		ctx->song = NULL;
		D ("\n");
		ctx->state = RHYTHMDB_TREE_PARSER_STATE_RHYTHMDB;
		break;
	}
	case RHYTHMDB_TREE_PARSER_STATE_ENTRY_PROPERTY:
	{
		switch (ctx->propid) {
		  
		case RHYTHMDB_PROP_TITLE:
			D ("%s => %s\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   ctx->buf->str);
			ctx->song->title = ctx->buf->str;
			g_string_free (ctx->buf, FALSE);
			break;

		case RHYTHMDB_PROP_GENRE:
			D ("%s => %s\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   ctx->buf->str);
			ctx->song->genre = ctx->buf->str;
			g_string_free (ctx->buf, FALSE);
			break;

		case RHYTHMDB_PROP_ARTIST:
			D ("%s => %s\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   ctx->buf->str);
			ctx->song->artist = ctx->buf->str;
			g_string_free (ctx->buf, FALSE);
			break;

		case RHYTHMDB_PROP_ALBUM:
			D ("%s => %s\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   ctx->buf->str);
			ctx->song->album = ctx->buf->str;
			g_string_free (ctx->buf, FALSE);
			break;

		case RHYTHMDB_PROP_LOCATION:
			D ("%s => %s\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   ctx->buf->str);
			ctx->song->pc_path_locale = ctx->buf->str;
			g_string_free (ctx->buf, FALSE);
			break;

		case RHYTHMDB_PROP_MIMETYPE:
			D ("Found mime type: %s\n", ctx->buf->str);
			/* FIXME: this assumes there's only one mime type tag
			 * for the current node
			 */
			ctx->add_song_to_db = ipod_is_mime_type_supported (ctx->buf->str);
			g_string_free (ctx->buf, TRUE);
			break;

		case RHYTHMDB_PROP_PLAY_COUNT: {
			gint val;
			val = strtoul (ctx->buf->str, NULL, 10);
			D ("%s => %d\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   val);
			g_string_free (ctx->buf, TRUE);
			ctx->song->playcount = val;
			break;
		}

		case RHYTHMDB_PROP_TRACK_NUMBER: {
			gint val;
			val = strtoul (ctx->buf->str, NULL, 10);
			D ("%s => %d\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   val);
			g_string_free (ctx->buf, TRUE);
			/* FIXME: make sure itunesdb properly support negative
			 * track numbers 
			 */
			ctx->song->track_nr = val;
			break;
		}

		case RHYTHMDB_PROP_DISC_NUMBER: {
			gint val;
			val = strtoul (ctx->buf->str, NULL, 10);
			D ("%s => %d\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   val);
			g_string_free (ctx->buf, TRUE);
			ctx->song->cd_nr = val;
			break;
		}

		case RHYTHMDB_PROP_BITRATE: {
			gint val;
			val = strtoul (ctx->buf->str, NULL, 10);
			D ("%s => %d\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   val);
			g_string_free (ctx->buf, TRUE);
			ctx->song->bitrate = val;
			break;
		}

		case RHYTHMDB_PROP_FILE_SIZE: {
			gint val;
			val = strtoul (ctx->buf->str, NULL, 10);
			D ("%s => %u\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   val);
			g_string_free (ctx->buf, TRUE);
			ctx->song->size = val;
			break;
		}

		case RHYTHMDB_PROP_MTIME: {
			/* glong */
			gint val;
			val = strtoul (ctx->buf->str, NULL, 10);
			D ("%s => %d\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   val);
			g_string_free (ctx->buf, TRUE);
			/* FIXME: properly encode mtime */
			break;
		}

		case RHYTHMDB_PROP_LAST_PLAYED: {
			/* glong */
			gint val;
			val = strtoul (ctx->buf->str, NULL, 10);
			D ("%s => %d\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   val);
			g_string_free (ctx->buf, TRUE);
			/* FIXME: properly encode last played time */
			break;
		}

		case RHYTHMDB_PROP_DURATION: {
			/* glong */
			gint val;
			val = strtoul (ctx->buf->str, NULL, 10);
			D ("%s => %d\n", 
			   entry_type_to_xml_tag (ctx->propid),
			   val);
			g_string_free (ctx->buf, TRUE);
			ctx->song->tracklen = val * 1000;
			break;
		}

		default:
			D ("Unsupported tag: %s\n", 
			   entry_type_to_xml_tag (ctx->propid));
			break;
		}

		ctx->state = RHYTHMDB_TREE_PARSER_STATE_ENTRY;
		break;
	}
	case RHYTHMDB_TREE_PARSER_STATE_START:
	case RHYTHMDB_TREE_PARSER_STATE_END:
	break;
	}
}

static void
rhythmdb_tree_parser_characters (struct RhythmDBTreeLoadContext *ctx, const char *data,
				 guint len)
{
	switch (ctx->state)
	{
	case RHYTHMDB_TREE_PARSER_STATE_ENTRY_PROPERTY:
		g_string_append_len (ctx->buf, data, len);
		break;
	case RHYTHMDB_TREE_PARSER_STATE_ENTRY:
	case RHYTHMDB_TREE_PARSER_STATE_RHYTHMDB:
	case RHYTHMDB_TREE_PARSER_STATE_START:
	case RHYTHMDB_TREE_PARSER_STATE_END:
	break;
	}
}

void
rhythmdb_tree_load (const gchar *name, iPodDB *db)
{
	xmlParserCtxtPtr ctxt;
	xmlSAXHandlerPtr sax_handler = g_new0 (xmlSAXHandler, 1);
	struct RhythmDBTreeLoadContext *ctx = g_new0 (struct RhythmDBTreeLoadContext, 1);

	sax_handler->startElement = (startElementSAXFunc) rhythmdb_tree_parser_start_element;
	sax_handler->endElement = (endElementSAXFunc) rhythmdb_tree_parser_end_element;
	sax_handler->characters = (charactersSAXFunc) rhythmdb_tree_parser_characters;

	ctx->state = RHYTHMDB_TREE_PARSER_STATE_START;
	ctx->db = db;

	if (g_file_test (name, G_FILE_TEST_EXISTS)) {
		ctxt = xmlCreateFileParserCtxt (name);
		ctx->xmlctx = ctxt;
		xmlFree (ctxt->sax);
		ctxt->userData = ctx;
		ctxt->sax = sax_handler;
		xmlParseDocument (ctxt);
		ctxt->sax = NULL;
		xmlFreeParserCtxt (ctxt);
	}
	g_free (sax_handler);
	g_free (ctx);
}
